package com.alibaba.cloud.security.sample.web;

import com.alibaba.cloud.security.sample.entity.UserDO;
import com.alibaba.cloud.security.sample.params.BaseResponse;
import com.alibaba.cloud.security.sample.params.simpleuser.*;
import com.alibaba.cloud.security.sample.service.IDaaSClientService;
import com.alibaba.cloud.security.sample.service.MockUserService;
import com.alibaba.cloud.security.sample.util.LoginUtil;
import com.aliyuncs.idaas_doraemon.model.v20210520.ListAuthenticatorsRequest;
import com.aliyuncs.idaas_doraemon.model.v20210520.ListAuthenticatorsResponse;
import com.aliyuncs.idaas_doraemon.model.v20210520.ServiceInvokeRequest;
import com.aliyuncs.idaas_doraemon.model.v20210520.ServiceInvokeResponse;
import com.google.gson.Gson;
import io.swagger.annotations.Api;
import io.swagger.annotations.ApiOperation;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.security.crypto.password.PasswordEncoder;
import org.springframework.stereotype.Controller;
import org.springframework.util.StringUtils;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestMethod;
import org.springframework.web.bind.annotation.ResponseBody;
import springfox.documentation.annotations.ApiIgnore;

import javax.servlet.http.HttpSession;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.util.Base64;
import java.util.HashMap;
import java.util.Map;

@Api
@Controller
@RequestMapping("/user")
public class SimpleUserController {

    @Value("${app.idaas.applicationExternalId}")
    String alibabaCloudIdaasApplicationExternalId;

    @Autowired
    PasswordEncoder passwordEncoder;

    @Autowired
    MockUserService mockUserService;
    @Autowired
    IDaaSClientService iDaaSClientService;

    @ApiOperation(value = "GetUserInfo", notes = "Get user info details")
    @ResponseBody
    @RequestMapping(value = "/info", method = RequestMethod.POST)
    public ResponseEntity<UserInfoResponse> info(@ApiIgnore HttpSession session) throws IOException {
        String userId = LoginUtil.getLoginUserId(session);
        UserDO user = null;
        if (StringUtils.hasText(userId)) {
            user = mockUserService.findUserByUserId(userId);
        }
        if (user == null) {
            return ResponseEntity.notFound().build();
        }
        UserInfoResponse userInfoResponse = new UserInfoResponse();
        userInfoResponse.setUserId(user.getUserId());
        userInfoResponse.setLoginId(user.getLoginId());
        userInfoResponse.setDisplayName(user.getDisplayName());
        return ResponseEntity.ok(userInfoResponse);
    }

    @ResponseBody
    @RequestMapping(value = "/create", method = RequestMethod.POST)
    public ResponseEntity<CreateUserResponse> create(@ApiIgnore HttpSession session, @RequestBody CreateUserRequest createUserRequest) throws IOException {
        // FIXME: 注意这里需要检测用户使用的密码是否发生过泄露，如泄露次数较多则不应当让用户使用该密码
        String loginId = createUserRequest.getLoginId();
        String displayName = createUserRequest.getDisplayName();
        String encodedPassword = passwordEncoder.encode(createUserRequest.getPassword());

        if (mockUserService.findUserByLoginId(loginId) != null) {
            CreateUserResponse userExistsResponse = new CreateUserResponse();
            userExistsResponse.setMessage("User exists.");
            return ResponseEntity.status(HttpStatus.BAD_REQUEST).body(userExistsResponse);
        }

        UserDO user = new UserDO();
        user.setLoginId(loginId);
        user.setPassword(encodedPassword);
        user.setDisplayName(displayName);
        String userId = mockUserService.createUser(user);

        LoginUtil.login(session, userId);
        return ResponseEntity.ok().build();
    }

    @ResponseBody
    @RequestMapping(value = "/query", method = RequestMethod.POST)
    public ResponseEntity<QueryUserResponse> query(@ApiIgnore HttpSession session, @RequestBody QueryUserRequest queryUserRequest) throws Exception {
        String loginId = queryUserRequest.getLoginId();

        UserDO user = mockUserService.findUserByLoginId(loginId);
        if (user == null) {
            QueryUserResponse queryUserResponse = new QueryUserResponse();
            queryUserResponse.setMessage("User not exists.");
            return ResponseEntity.status(HttpStatus.NOT_FOUND).body(queryUserResponse);
        }

        QueryUserResponse queryUserResponse = new QueryUserResponse();
        queryUserResponse.setUserId(user.getUserId());

        // 检查是否已经注册WebAuthn
        ListAuthenticatorsRequest listAuthenticatorsRequest = new ListAuthenticatorsRequest();
        listAuthenticatorsRequest.setApplicationExternalId(alibabaCloudIdaasApplicationExternalId);
        listAuthenticatorsRequest.setUserId(user.getUserId());
        listAuthenticatorsRequest.setPageSize(100L);
        ListAuthenticatorsResponse listAuthenticatorsResponse = iDaaSClientService.getClient().getAcsResponse(listAuthenticatorsRequest);
        boolean hasRegisteredWebAuthn = false;
        boolean isOpenTotp = false;
        if (listAuthenticatorsResponse.getAuthenticator() != null) {
            for (ListAuthenticatorsResponse.AuthenticatorListDTO authenticatorListDTO : listAuthenticatorsResponse.getAuthenticator()) {
                if ("WEBAUTHN".equals(authenticatorListDTO.getType())) {
                    hasRegisteredWebAuthn = true;
                } else if ("TOTP".equals(authenticatorListDTO.getType())) {
                    isOpenTotp = true;
                }
            }
        }
        queryUserResponse.setRegisteredWebAuthn(hasRegisteredWebAuthn);
        queryUserResponse.setOpenTotp(isOpenTotp);

        return ResponseEntity.ok(queryUserResponse);
    }

    @ResponseBody
    @RequestMapping(value = "/login", method = RequestMethod.POST)
    public ResponseEntity<LoginUserResponse> login(@ApiIgnore HttpSession session, @RequestBody LoginUserRequest loginUserRequest) throws Exception {
        // FIXME: 注意这里需要进行频率及风险控制，不然会有安全风险，本程序是常规程序所以没有做该控制
        // 另外，如果检测到用户当前使用的密码泄露应当引进行两步验证等其他再次验证措施
        String loginId = loginUserRequest.getLoginId();
        UserDO user = mockUserService.findUserByLoginId(loginId);
        if (user == null) {
            LoginUserResponse loginUserResponse = new LoginUserResponse();
            loginUserResponse.setMessage("User not exists, TOTP or password error.");
            return ResponseEntity.status(HttpStatus.NOT_FOUND).body(loginUserResponse);
        }

        // 检查用户是否开启了TOTP
        ListAuthenticatorsRequest listAuthenticatorsRequest = new ListAuthenticatorsRequest();
        listAuthenticatorsRequest.setApplicationExternalId(alibabaCloudIdaasApplicationExternalId);
        listAuthenticatorsRequest.setUserId(user.getUserId());
        listAuthenticatorsRequest.setAuthenticatorType("TOTP");
        listAuthenticatorsRequest.setPageSize(2L);
        ListAuthenticatorsResponse listAuthenticatorsResponse = iDaaSClientService.getClient().getAcsResponse(listAuthenticatorsRequest);
        boolean isOpenTotp = listAuthenticatorsResponse.getAuthenticator() != null && listAuthenticatorsResponse.getAuthenticator().size() > 0;

        if (isOpenTotp && !StringUtils.hasText(loginUserRequest.getTotpCode())) {
            LoginUserResponse loginUserResponse = new LoginUserResponse();
            loginUserResponse.setMessage("TOTP is required");
            return ResponseEntity.badRequest().body(loginUserResponse);
        }

        if (passwordEncoder.matches(loginUserRequest.getPassword(), user.getPassword())) {
            if (isOpenTotp) {
                Map<String, Object> otpVerifyParams = new HashMap<>();
                otpVerifyParams.put("uniqueId", user.getUserId());
                otpVerifyParams.put("uniqueIdType", "USER_ID");
                otpVerifyParams.put("code", loginUserRequest.getTotpCode());

                ServiceInvokeRequest serviceInvokeRequest = new ServiceInvokeRequest();
                serviceInvokeRequest.setApplicationExternalId(alibabaCloudIdaasApplicationExternalId);
                serviceInvokeRequest.setServiceCode("TOTP");
                serviceInvokeRequest.setDoraemonAction("VerifyOTP");
                serviceInvokeRequest.setServerExtendParamsJson(Base64.getEncoder().encodeToString(
                        new Gson().toJson(otpVerifyParams).getBytes(StandardCharsets.UTF_8)
                ));
                serviceInvokeRequest.setMobileExtendParamsJson("");
                serviceInvokeRequest.setMobileExtendParamsJsonSign("");

                boolean verifyTotpResult;
                try {
                    ServiceInvokeResponse response = iDaaSClientService.getClient().getAcsResponse(serviceInvokeRequest);
                    verifyTotpResult = "Operation.Success".equals(response.getCode());
                } catch (com.aliyuncs.exceptions.ServerException e) {
                    if ("Operation.Failure.Otp.VerifyErrorForCodeWrong".equals(e.getErrCode())) {
                        verifyTotpResult = false;
                    } else { // rethrow
                        throw e;
                    }
                }

                if (!verifyTotpResult) {
                    LoginUserResponse loginUserResponse = new LoginUserResponse();
                    loginUserResponse.setMessage("User not exists, TOTP or password error...");
                    return ResponseEntity.badRequest().body(loginUserResponse);
                }
            }

            LoginUserResponse loginUserResponse = new LoginUserResponse();
            LoginUtil.login(session, user.getUserId());
            return ResponseEntity.ok(loginUserResponse);
        } else {
            LoginUserResponse loginUserResponse = new LoginUserResponse();
            loginUserResponse.setMessage("User not exists, TOTP or password error..");
            return ResponseEntity.badRequest().body(loginUserResponse);
        }
    }

    @ApiOperation(value = "LogoutUser", notes = "Make user logout")
    @ResponseBody
    @RequestMapping(value = "/logout", method = RequestMethod.POST)
    public ResponseEntity<BaseResponse> logout(@ApiIgnore HttpSession session) throws IOException {
        LoginUtil.logout(session);
        return ResponseEntity.ok().build();
    }
}
